
#import "ObjectDetector.h"
#import "CvInvoke.h"
#import "EngineUtils.h"
#import "MatManager.h"

static CvVideoCamera *videoCamera;

@implementation ObjectDetector {
    // private properties
    double mCurrentMillis;
    bool mTakePicture;
    bool framesInitialized;
    
    // recording
    cv::VideoWriter mVideoWriter;
    int mWidth;
    int mHeight;
    
    Mat mSampleImageMat;
    Scalar mOverlayColor;
    float mOverlayTransparency;
    
    bool mDetecting;
    Ptr<cv::ORB> mORB;
    Ptr<cv::DescriptorMatcher>      mDescriptorMatcher;
    std::vector<cv::KeyPoint>   msPoints;
    Mat                    msDes;
    Mat                    mWarpMatrix;
    Mat                    mMask;
    int                    mNDetected;
}

- (id)initWithBridge:(RCTBridge *)bridge
{
    if ((self = [super init])) {
        self.startDate = [NSDate date];
        self.mOverlayInterval = 0;
        mCurrentMillis = 0;
        self.backgroundColor = [UIColor blackColor];
        self.bridge = bridge;
        
        mOverlayColor = cv::Scalar(0, 0, 255);
        mOverlayTransparency = 0.2f;
        
        mORB = cv::ORB::create(15000, 1.3f, 8, 10, 1, 2, ORB::HARRIS_SCORE, 31, 10);
        mDescriptorMatcher = cv::BFMatcher::create(cv::BFMatcher::BRUTEFORCE_HAMMING, true);
        self.mNDetectionB4Take = 1;
        
        [[NSNotificationCenter defaultCenter] addObserver:self
                                                 selector:@selector(bridgeDidBackground:)
                                                     name:UIApplicationDidEnterBackgroundNotification
                                                   object:nil];
        [[NSNotificationCenter defaultCenter] addObserver:self
                                                 selector:@selector(bridgeDidForeground:)
                                                     name:UIApplicationWillEnterForegroundNotification
                                                   object:nil];
        
        if (videoCamera != nil) {
            // if more than one camera instance is used there are memory issues ...
            //[videoCamera stop];
            [videoCamera setDelegate:self];
            [videoCamera setParentView:self];
            //[videoCamera start];
        }
        else {
            videoCamera = [[CvVideoCamera alloc] initWithParentView:self];
            videoCamera.defaultAVCaptureDevicePosition = AVCaptureDevicePositionBack;
            videoCamera.defaultAVCaptureSessionPreset = AVCaptureSessionPreset1280x720;
            videoCamera.defaultAVCaptureVideoOrientation = AVCaptureVideoOrientationPortrait;
            videoCamera.defaultFPS = 30;
            videoCamera.grayscaleMode = false;
            videoCamera.delegate = self;
            //self.videoCamera = videoCamera;
            [videoCamera start];
        }
    }
    [self initParams];
    
    return self;
}

/** TODO: implement notifcations so these methods get called */
- (void)bridgeDidForeground:(NSNotification *)notification {
    [self initParams];
    [videoCamera start];
}

- (void)bridgeDidBackground:(NSNotification *)notification {
    [videoCamera stop];
}

- (void)initParams {
    mNDetected = 0;
    mDetecting = false;
}

- (void) getFeaturesOfSample {
    cv::resize(mSampleImageMat, mSampleImageMat, cv::Size(250, 250));
    
    mORB->setFirstLevel(0);
    mORB->detectAndCompute(mSampleImageMat, cv::noArray(), msPoints, msDes);
    mORB->setFirstLevel(1);
}

- (cv::Mat)rotateMat:(cv::Mat) image withAngle:(int) rotFlag {
    Mat rotated;
    //1=CW, 2=CCW, 3=180
    switch (rotFlag) {
        case 1:
            cv::transpose(image, rotated);
            cv::flip(rotated, rotated, 1);
            return rotated;
        case 2:
            cv::transpose(image, rotated);
            cv::flip(rotated, rotated, 0);
            return rotated;
        case 3:
            cv::flip(image, rotated, -1);
            return rotated;
    }
    
    return image;
}

- (void)setSampleImage:(NSString *)uri {
    NSURL *url = [NSURL URLWithString:uri];
    
    if (url && url.scheme && url.host) {
        dispatch_async(dispatch_queue_create("image_processing", 0), ^{
           NSData* data = [NSData dataWithContentsOfURL:url];
           CGImageSourceRef source = CGImageSourceCreateWithData((CFMutableDataRef)data, NULL);
           NSDictionary *metadata = (NSDictionary *)CFBridgingRelease(CGImageSourceCopyPropertiesAtIndex(source,0,NULL));
            
           NSNumber *orientation = [metadata valueForKey:@"Orientation"];
            
           dispatch_async(dispatch_get_main_queue(), ^{
               cv::Mat matC = cv::imdecode(Mat(1, (int)[data length], CV_8UC1, (void*)data.bytes), CV_LOAD_IMAGE_UNCHANGED);
               matC = [self rotateMat:matC withAngle:[orientation intValue]];
               
               cv::cvtColor(matC, self->mSampleImageMat, cv::COLOR_RGB2GRAY);
               [self getFeaturesOfSample];
           });
        });
    } else {
        @try {
            if (uri.length > 0 && [uri hasPrefix:@"file://"]) {
                const char * cpath = [[uri substringFromIndex:7] cStringUsingEncoding:NSUTF8StringEncoding];
                // NSURL *myURL = [NSURL fileURLWithPath:[NSString stringWithUTF8String:cpath]];
                
                // CGImageSourceRef source = CGImageSourceCreateWithURL((CFURLRef)myURL, NULL);
                // NSDictionary *metadata = (NSDictionary *)CFBridgingRelease(CGImageSourceCopyPropertiesAtIndex(source,0,NULL));
                 
                // NSNumber *orientation = [metadata valueForKey:@"Orientation"];
                
                mSampleImageMat = cv::imread(cpath, CV_LOAD_IMAGE_GRAYSCALE);
                
                /*
                if (!mSampleImageMat.empty()) {
                    UIImage *img = MatToUIImage(mSampleImageMat);
                    UIImageWriteToSavedPhotosAlbum(img, self, nil, nil);
                }
                */
                
                // mSampleImageMat = [self rotateMat:mSampleImageMat withAngle:[orientation intValue]];
                [self getFeaturesOfSample];
            }
        }
        @catch (NSException *e) {
            
        }
    }
}

- (void)setOverlayColor:(NSString *)hexString {
    NSString *str = [hexString stringByReplacingOccurrencesOfString:@"#" withString:@""];
    unsigned r,g,b,a;
    
    switch(str.length) {
        case 6:
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(0, 2)]] scanHexInt:&r];
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(2, 2)]] scanHexInt:&g];
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(4, 2)]] scanHexInt:&b];
            
            mOverlayColor = cv::Scalar(r, g, b);
            mOverlayTransparency = 1.0f;
            break;
        case 8:
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(0, 2)]] scanHexInt:&r];
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(2, 2)]] scanHexInt:&g];
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(4, 2)]] scanHexInt:&b];
            [[NSScanner scannerWithString:[str substringWithRange:NSMakeRange(6, 2)]] scanHexInt:&a];

            mOverlayColor = cv::Scalar(r, g, b);
            mOverlayTransparency = a / 255.0f;
            break;
    }
}


- (NSString*) toBase64String:(cv::Mat)image {
    std::vector<uchar> buf;
    cv::imencode(cv::String(".jpg"), image, buf);
    NSData *data = [NSData dataWithBytes:buf.data() length:buf.size()];
    return [data base64EncodedStringWithOptions:0];
}

- (BOOL)isNiceHomography:(cv::Mat)H {
    double a00 = H.at<double>(0, 0);
    double a01 = H.at<double>(0, 1);
    double a10 = H.at<double>(1, 0);
    double a11 = H.at<double>(1, 1);
    double a22 = H.at<double>(2, 2);

    double xNorm = a00*a00 + a01*a01;
    double yNorm = a10*a10 + a11*a11;
    double scal = a00*a10 + a01*a11;

    return xNorm > 0.04 && xNorm < 9
            && yNorm > 0.04 && yNorm < 9
            && scal > -0.5 && scal < 0.5;
}

- (void)processImage:(cv::Mat&)image {

    // Do some OpenCV stuff with the image
    Mat gray;
    cv::Mat image_copy; // image in corrected colorspace ...
    cvtColor(image, image_copy, COLOR_BGR2RGBA);
    cvtColor(image, gray, COLOR_BGR2GRAY);
    
    mWidth = image.cols;
    mHeight = image.rows;
    
    if (mMask.empty() || mMask.size != image.size) {
        mMask = cv::Mat(image.rows, image.cols, CV_8U, Scalar::all(0));
        cv::rectangle(mMask, cv::Point(image.cols / 6, image.rows / 6), cv::Point(image.cols * 5 / 6, image.rows * 5 / 6), Scalar(255), -1);
    }
    
    //Log.d(TAG, "payload is: " + faceInfo);
    if (self && self.onFrameSize) {
        NSString *frameSizeStr = [NSString stringWithFormat:@"{\"frameSize\":{\"frameWidth\":%d,\"frameHeight\":%d}}",image.cols,image.rows];
        self.onFrameSize(@{@"payload":frameSizeStr});
    }
    
    if (!self->mDetecting && self->msPoints.size() >= 10) {
        self->mDetecting = true;
        
        dispatch_queue_t queue = dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0);
        dispatch_async(queue, ^{
            cv::GaussianBlur(gray, gray, cv::Size(5,5), 0);
            std::vector<cv::KeyPoint> kp;
            Mat des;
            self->mORB->detectAndCompute(gray, self->mMask, kp, des);
            
            Mat warp, emptyMat;
            
            if (kp.size() >= 10) {
                std::vector<cv::DMatch> matches;
                self->mDescriptorMatcher->match(Mat(self->msDes), Mat(des), matches);
                
                if (matches.size() > 3) {
                    std::vector<Point2f> sPoints, dPoints;
                    for (int i=0; i<matches.size(); i++) {
                    sPoints.push_back(self->msPoints[matches[i].queryIdx].pt);
                        dPoints.push_back(kp[matches[i].trainIdx].pt);
                    }
                    
                    warp = cv::findHomography(sPoints, dPoints, cv::RANSAC, 5.0f);
                    if (!warp.empty() && ![self isNiceHomography:warp]) {
                        warp = emptyMat;
                    }
                }
            }


            dispatch_async(dispatch_get_main_queue(), ^{
                //tell the main UI thread here
                self->mWarpMatrix = warp;
                if (!warp.empty()) {
                    self->mNDetected++;
                    if (self->mNDetected >= self.mNDetectionB4Take) {
                        if (self && self.onObjectCaptured) {
                            Mat invWarp = warp.inv();
                            Mat warpped;
                            cv::warpPerspective(image_copy, warpped, invWarp, cv::Size(self->mSampleImageMat.rows, self->mSampleImageMat.cols));
                            cv::resize(warpped, warpped, cv::Size(500, 500));
                            cv::cvtColor(warpped, warpped, cv::COLOR_BGR2RGB);
                            
                            NSString *base64Warpped = [self toBase64String:warpped];
                            self.onObjectCaptured(@{@"payload": base64Warpped});
                        }
                        self->mNDetected = 0;
                    }
                }
                self->mDetecting = false;
            });
        });
    }

    dispatch_async(dispatch_get_main_queue(), ^{
    
        if (self.mCvInvokeGroup != nil) {
            double currMillis = [self.startDate timeIntervalSinceNow];
            double diff = (currMillis - self->mCurrentMillis) * 1000.0;
            diff = (diff < 0.0) ? -diff : diff;
            if (diff >= (double)[self.mOverlayInterval doubleValue]) {
                
                self->mCurrentMillis = currMillis;
                self.startDate = [NSDate date];
                
                CvInvoke *invoker = [[CvInvoke alloc] initWithRgba:image_copy gray:gray];
                NSArray *responseArr = [invoker parseInvokeMap:self.mCvInvokeGroup];
                NSString *lastCall = invoker.callback;
                int dstMatIndex = invoker.dstMatIndex;
                
                if (lastCall != nil && lastCall != (NSString*)NSNull.null && ![lastCall isEqualToString:@""] && dstMatIndex >= 0 && dstMatIndex < 1000) {
                    if (responseArr && responseArr.count > 0 && self && self.onPayload) {
                        self.onPayload(@{ @"payload" : responseArr });
                    }
                }
                else {
                    // not necessarily error condition unless dstMatIndex >= 1000
                    if (dstMatIndex >= 1000) {
                        NSLog(@"Exception thrown attempting to invoke method.  Check your method name and parameters and make sure they are correct.");
                    }
                }
            }
        }
    
        if (self.mOverlay != nil) {
            int matIndex = [[self.mOverlay valueForKey:@"matIndex"] intValue];
            Mat overlayMat = [MatManager.sharedMgr matAtIndex:matIndex];
            addWeighted(image_copy, 1.0, overlayMat, 1.0, 0.0, image_copy);
        }
    
        if (self->mTakePicture) {
            self->mTakePicture = false;
            MatWrapper *MW = [[MatWrapper alloc] init];
            MW.myMat = image_copy;
            self.takePicBlock(MW);
        }
        
        UIImage *videoImage;
        if (!self->mWarpMatrix.empty()) {
            std::vector<Point2f> obj_corners, scene_corners;
            obj_corners.push_back(Point2f(0,0));
            obj_corners.push_back(Point2f(self->mSampleImageMat.cols,0));
            obj_corners.push_back(Point2f(self->mSampleImageMat.cols, self->mSampleImageMat.rows));
            obj_corners.push_back(Point2f(0, self->mSampleImageMat.rows));
            
            cv::perspectiveTransform(obj_corners, scene_corners, self->mWarpMatrix);
            
            cv::Point scene_points[1][20];
            scene_points[0][0] = cv::Point( scene_corners[0].x, scene_corners[0].y );
            scene_points[0][1] = cv::Point( scene_corners[1].x, scene_corners[1].y );
            scene_points[0][2] = cv::Point( scene_corners[2].x, scene_corners[2].y );
            scene_points[0][3] = cv::Point( scene_corners[3].x, scene_corners[3].y );

            const cv::Point* ppt[1] = { scene_points[0] };
            int npt[] = { 4 };

            Mat mask = image_copy.clone();
            cv::fillPoly(mask, ppt, npt, 1, self->mOverlayColor);
            cv::addWeighted(image_copy, (1 - self->mOverlayTransparency), mask, self->mOverlayTransparency, 0, mask);
            videoImage = MatToUIImage(mask);
        } else {
            videoImage = MatToUIImage(image_copy);
        }
        
    
        [self setImage:videoImage];
    });
}

#pragma Properties

- (void)changeFacing:(NSString*)facing {
    if (![facing isEqualToString:self.mFacing]) {
        self.mFacing = facing;
        [videoCamera stop];
        if ([self.mFacing isEqualToString:@"back"]) {
            [videoCamera setDefaultAVCaptureDevicePosition:AVCaptureDevicePositionBack];
        }
        else {
            [videoCamera setDefaultAVCaptureDevicePosition:AVCaptureDevicePositionFront];
        }
        [videoCamera start];
    }
}

- (void)takePicture:(TakePicBlock)takePicBlock {
    mTakePicture = true;
    self.takePicBlock = takePicBlock;
}

@end
